#define _GNU_SOURCE

#include "append_list.h"
#include "ch_time_and_mod_dir.h"
#include "compare_verify_one.h"
#include "create_symboliclink.h"
#include "print_error.h"
#include "read_write.h"
#include "remove_source_rmdir.h"
#include "remove_source_unlink.h"
#include "struct_CPDD.h"
#include "struct_CPInfo.h"
#include "struct_DPath.h"
#include "struct_OPArg.h"
#include "struct_RData.h"
#include "struct_SDir.h"
#include "xmalloc.h"

#include <limits.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <linux/limits.h>

/* 関数プロトタイプ */
RData * check_target(const OPArg *, const VList *drive_list, const SDir *);
static inline VList * copy_arg_link(const OPArg *, const SDir *, CPInfo *, CPDD *, VList *file_list);
static inline VList * copy_arg_file(const VList *drive_list, const char src[], const char dst[], const OPArg *, const SDir *, CPInfo *, CPDD *, VList *file_list);
static inline VList * copy_arg_dir(const char *tmp_src, const char *tmp_dst, const _Bool is_src_current, const VList *drive_list, const OPArg *, const SDir *, CPInfo *, CPDD *, VList *file_list);
static void check_overwrite(const OPArg *, const SDir *, CPInfo *, CPDD *, const char type);
static void check_write_file_or_create_link(const OPArg *, const SDir *, CPInfo *, CPDD *);
static inline _Bool check_parent(const char *src, const char *dst);
static inline struct stat * make_directory(const OPArg *, CPInfo *, CPDD *);
static VList * dir_traverse(const OPArg *, const SDir *, VList *file_list, CPInfo *, CPDD *);
static inline void delete_last_slash(const int len, char []);
static inline CPInfo * malloc_cpinfo(void);
static inline char * get_dst_plus_srcname(const char *src, const char *dst);
static inline void count_plus_link(CPInfo *, CPDD *);
static inline void count_plus_file(CPInfo *, CPDD *);
static inline void count_plus_dir(CPInfo *, CPDD *);
static inline void clean(const char buf[], const int len);
static inline void fgets_is_null(void);
static inline void set_same_or_different(const VList *drive_list, const char *src, const char *dst, const OPArg *, CPInfo *);

/*******************************************************************************
*******************************************************************************/
RData * check_target(const OPArg *oparg, const VList *drive_list, const SDir *sdir)
{
	int loop_dst = oparg->dst_index;

	if(g_file_test(oparg->argv[loop_dst], G_FILE_TEST_IS_SYMLINK) == TRUE)
	{
		fprintf(stderr, "コピー先がシンボリックリンクです\n");
		exit(EXIT_FAILURE);
	}

	enum argument_count
	{
		SINGULAR,
		PLURAL,
	} acount;

	int loop = oparg->src_index;

	/* 三項演算子 */
	((loop_dst - loop) == 1) ? (acount = SINGULAR) : (acount = PLURAL);

	/*
	 * CPDD *cpdd = xmalloc0(sizeof(CPDD));
	 * ↑のコードだと構造体のメンバが0で初期化されるとは限らないらしい。
	 * 大抵の処理系では期待通りに動くらしいけど。
	 */
	CPDD *cpdd = xmalloc(sizeof(CPDD));
	cpdd->file_count = 0;
	cpdd->dir_count = 0;
	cpdd->link_count = 0;
	cpdd->mk_file_count = 0;
	cpdd->mk_dir_count = 0;
	cpdd->mk_link_count = 0;
	cpdd->total_source = 0;
	cpdd->total_read = 0;
	cpdd->total_write = 0;
	cpdd->total_error = 0;
	cpdd->thread_mode_count = 0;

	/* コピー元が単一かつ . で、コピー先が存在しないとき用の処理 */
	{
		char *s = oparg->argv[loop];

		if((acount == SINGULAR) && (*s == '.'))
		{
			_Bool is_current = true;
			s++;
			char t = G_DIR_SEPARATOR;

			/* コピー元が . かどうかチェック。 ././. とかの場合は . に入れ替える。 */
			for(;;)
			{
				if(*s == t)
				{
					s++;
					if(t == '.')
					{
						t = G_DIR_SEPARATOR;
					}
					else
					{
						t = '.';
					}
				}
				else if(*s == '\0')
				{
					oparg->argv[loop] = ".";
					break;
				}
				else
				{
					is_current = false;
					break;
				}
			}

			if(is_current == true)
			{
				errno = 0;

				if(mkdir(oparg->argv[loop_dst], 0755) == -1)
				{
					if((errno != EEXIST) && (errno != 0))
					{
						print_error("mkdir", __FILE__, __LINE__, cpdd);
						fprintf(stderr, "error : %s\n", oparg->argv[loop_dst]);
						exit(EXIT_FAILURE);
					}
				}
			}
		}
	}

	char tmp_dst[PATH_MAX];
	_Bool file_found = false;
	errno = 0;
	realpath(oparg->argv[loop_dst], tmp_dst);

	if(errno == 0)
	{
		file_found = true;
	}
	else if(errno == ENOENT)
	{
		char *d = oparg->argv[loop_dst];
		int dlen = strlen(d);
		char *c;

		/* 末尾がパス区切り文字の場合はヌル文字に変えておく */
		if(d[dlen - 1] == G_DIR_SEPARATOR)
		{
			char *z = xmalloc(dlen + 1);
			strcpy(z, d);
			delete_last_slash(dlen, z);
			c = g_path_get_dirname(z);
			oparg->argv[loop_dst] = z;
		}
		else
		{
			c = g_path_get_dirname(oparg->argv[loop_dst]);
		}

		errno = 0;
		realpath(c, tmp_dst);

		if(errno != 0)
		{
			print_error("realpath", __FILE__, __LINE__, cpdd);
			fprintf(stderr, "error : %s\n", oparg->argv[loop_dst]);
			exit(EXIT_FAILURE);
		}

		free(c);
	}
	else
	{
		print_error("realpath", __FILE__, __LINE__, cpdd);
		fprintf(stderr, "error : %s\n", oparg->argv[loop_dst]);
		exit(EXIT_FAILURE);
	}

	VList *file_list = NULL;
	char tmp_src[PATH_MAX];
	int len_dst = strlen(tmp_dst);
	delete_last_slash(len_dst, tmp_dst);

	for(; loop < loop_dst; loop++)
	{
		_Bool is_src_current = false;

		/* strcmpは等しければ0を返す */
		if(strcmp(oparg->argv[loop], "..") == 0)
		{
			fprintf(stderr, "skip : ..\n");
			continue;
		}
		else if(strcmp(oparg->argv[loop], ".") == 0)
		{
			is_src_current = true;
		}
		/*
		 * メモ
		 * cpコマンドの場合、引数が.だったらそのフォルダにあるファイルをコピーする。
		 * 引数が..だとエラー。
		 */

		_Bool is_link = false;

		if(g_file_test(oparg->argv[loop], G_FILE_TEST_IS_SYMLINK) == TRUE)
		{
			is_link = true;
			strcpy(tmp_src, oparg->argv[loop]);
		}
		else
		{
			errno = 0;
			realpath(oparg->argv[loop], tmp_src);

			if(errno != 0)
			{
				print_error("realpath", __FILE__, __LINE__, cpdd);
				fprintf(stderr, "error : %s\n", oparg->argv[loop]);
				continue;
			}
		}

		int len_src = strlen(tmp_src);
		delete_last_slash(len_src, tmp_src);

		CPInfo *cpinfo = malloc_cpinfo();

		len_src++;
		cpinfo->src = xmalloc(len_src);
		memcpy(cpinfo->src, tmp_src, len_src);
		len_src--;

		switch(file_found)
		{
		case true:
			if(g_file_test(tmp_dst, G_FILE_TEST_IS_REGULAR) == TRUE)
			{
				len_dst++;
				cpinfo->dst = xmalloc(len_dst);
				memcpy(cpinfo->dst, tmp_dst, len_dst);
				len_dst--;

				switch(acount)
				{
				case SINGULAR:
					if(is_link == true)
					{
						fprintf(stderr, "リンク (%s) に対するコピー先が不正 (ファイル) です\n", tmp_src);
						exit(EXIT_FAILURE);
					}
					else if(g_file_test(tmp_src, G_FILE_TEST_IS_REGULAR) == TRUE)
					{
						file_list = copy_arg_file(drive_list, tmp_src, tmp_dst, oparg, sdir, cpinfo, cpdd, file_list);
					}
					else if(g_file_test(tmp_src, G_FILE_TEST_IS_DIR) == TRUE)
					{
						fprintf(stderr, "フォルダ (%s) に対するコピー先が不正 (ファイル) です\n", tmp_src);
						exit(EXIT_FAILURE);
					}
					else
					{
						fprintf(stderr, "コピー先が不正 (ファイル) です\n");
						exit(EXIT_FAILURE);
					}
					break;

				default:
					fprintf(stderr, "複数のコピー元に対するコピー先が不正 (ファイル) です\n");
					exit(EXIT_FAILURE);
					break;
				}
			}
			else if(g_file_test(tmp_dst, G_FILE_TEST_IS_DIR) == TRUE)
			{
				cpinfo->dst = get_dst_plus_srcname(tmp_src, tmp_dst);

				if(is_link == true)
				{
					file_list = copy_arg_link(oparg, sdir, cpinfo, cpdd, file_list);
				}
				else if(g_file_test(tmp_src, G_FILE_TEST_IS_REGULAR) == TRUE)
				{
					file_list = copy_arg_file(drive_list, tmp_src, tmp_dst, oparg, sdir, cpinfo, cpdd, file_list);
				}
				else if(g_file_test(tmp_src, G_FILE_TEST_IS_DIR) == TRUE)
				{
					file_list = copy_arg_dir(tmp_src, tmp_dst, is_src_current, drive_list, oparg, sdir, cpinfo, cpdd, file_list);
				}
				else
				{
					fprintf(stderr, "skip : %s\n", tmp_src);
				}
			}
			else
			{
				fprintf(stderr, "コピー先が不正です\n");
				exit(EXIT_FAILURE);
			}
			break;

		case false:
			if(acount == PLURAL)
			{
				errno = 0;
				if(mkdir(oparg->argv[loop_dst], 0755) == -1)
				{
					if((errno != EEXIST) && (errno != 0))
					{
						print_error("mkdir", __FILE__, __LINE__, cpdd);
						fprintf(stderr, "error : %s\n", oparg->argv[loop_dst]);
						exit(EXIT_FAILURE);
					}
				}
				else
				{
					char *target = get_dst_plus_srcname(oparg->argv[loop_dst], tmp_dst);
					strcpy(tmp_dst, target);
					cpinfo->dst = get_dst_plus_srcname(tmp_src, target);
					free(target);
					file_found = true;
				}
			}
			else
			{
				cpinfo->dst = get_dst_plus_srcname(oparg->argv[loop_dst], tmp_dst);
			}

			if(is_link == true)
			{
				file_list = copy_arg_link(oparg, sdir, cpinfo, cpdd, file_list);
			}
			else if(g_file_test(tmp_src, G_FILE_TEST_IS_REGULAR) == TRUE)
			{
				file_list = copy_arg_file(drive_list, tmp_src, tmp_dst, oparg, sdir, cpinfo, cpdd, file_list);
			}
			else if(g_file_test(tmp_src, G_FILE_TEST_IS_DIR) == TRUE)
			{
				file_list = copy_arg_dir(tmp_src, tmp_dst, is_src_current, drive_list, oparg, sdir, cpinfo, cpdd, file_list);
			}
			else
			{
				fprintf(stderr, "skip : %s\n", tmp_src);
			}
			break;
		}
	}

	RData *rdata = xmalloc(sizeof(RData));
	rdata->file_list = file_list;
	rdata->cpdd = cpdd;

/*
	FREE_THIS_DIR
*/
	return rdata;
}

/*******************************************************************************
*******************************************************************************/
static inline VList * copy_arg_link(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd, VList *file_list)
{
	count_plus_link(cpinfo, cpdd);
	check_overwrite(oparg, sdir, cpinfo, cpdd, 'f');
	file_list = append_list((void *)cpinfo, file_list);
	return file_list;
}

/*******************************************************************************
*******************************************************************************/
static inline VList * copy_arg_file(const VList *drive_list, const char tmp_src[], const char tmp_dst[], const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd, VList *file_list)
{
	count_plus_file(cpinfo, cpdd);
	set_same_or_different(drive_list, tmp_src, tmp_dst, oparg, cpinfo);
	check_overwrite(oparg, sdir, cpinfo, cpdd, 'f');
	file_list = append_list((void *)cpinfo, file_list);
	return file_list;
}

/*******************************************************************************
*******************************************************************************/
static inline VList * copy_arg_dir(const char *tmp_src, const char *tmp_dst, const _Bool is_src_current, const VList *drive_list, const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd, VList *file_list)
{
	_Bool is_parent = check_parent(tmp_src, tmp_dst);

	if(is_parent == true)
	{
		if(is_src_current == true)
		{
			fprintf(stderr, "skip : .\n");
		}
		else
		{
			fprintf(stderr, "skip : %s\n", cpinfo->src);
		}

		if(cpinfo->src != NULL) free(cpinfo->src);
		if(cpinfo->dst != NULL) free(cpinfo->dst);

		free(cpinfo);
		cpinfo = NULL;
	}
	else
	{
		set_same_or_different(drive_list, tmp_src, tmp_dst, oparg, cpinfo);

		struct stat *stat_buf;

		if(is_src_current == false)
		{
			count_plus_dir(cpinfo, cpdd);
			stat_buf = make_directory(oparg, cpinfo, cpdd);
			file_list = append_list((void *)cpinfo, file_list);
		}
		else
		{
			char *current_tmp = g_path_get_dirname(cpinfo->dst);
			free(cpinfo->dst);
			cpinfo->dst = current_tmp;
			/* cpinfoをリストに追加していないので、メモリリークしている。 */
		}

		file_list = dir_traverse(oparg, sdir, file_list, cpinfo, cpdd);

		if((is_src_current == false) && (cpinfo->dst_para == WRITE))
		{
			ch_time_and_mod_dir(cpinfo, stat_buf, cpdd);
		}

		if((oparg->M == MOVE) && ((oparg->CHECK == NOT) || (oparg->CHECK_MODE == BEFORE))) { REMOVE_SOURCE_RMDIR }
	}

	return file_list;
}

/*******************************************************************************
*******************************************************************************/
#define MAKE_THIS_DIR { this_dir = g_path_get_dirname(cpinfo->dst); }
#define FREE_THIS_DIR \
{\
	if(this_dir != NULL)\
	{\
		free(this_dir);\
		this_dir = NULL;\
	}\
}
#define DO_NOT_OVERWRITE { cpinfo->dst_para = SKIP; }

/*******************************************************************************
*******************************************************************************/
static void check_overwrite(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd, const char type)
{
	enum I_OverWrite_Type
	{
		INFO,
		THIS_DIR_YES,
		THIS_DIR_NO,
	};

	static _Bool is_info;
	static enum I_OverWrite_Type iowt;
	static OPFlag I;
	static OPFlag WMODE;

	if(is_info == false)
	{
		is_info = true;
		iowt = INFO;

		/* 変更される可能性があるオプションはローカル変数に入れておく */
		I = oparg->I;
		WMODE = oparg->WRITE_MODE;
	}

	gboolean fcheck  = g_file_test(cpinfo->dst, G_FILE_TEST_EXISTS);

	/* TRUE・FALSE　→　glib */
	/* true・false　→　stdbool */
	if(fcheck == FALSE)
	{
		check_write_file_or_create_link(oparg, sdir, cpinfo, cpdd);
	}
	else
	{
		cpinfo->file_test = true;

		if(I == INTERACTIVE)
		{
			/*
			 * -i（上書き確認）モード
			 * ここの処理はgoto使わずに関数にまとめた方が分かりやすいかもしれない
			 */
INPUT:
			/*
			 * gotoのラベルの直後で変数の宣言をすると、
			 * a label can only be part of a statement and a declaration is not a statement
			 * というエラーが出る。
			 */

			switch(iowt)
			{
			case INFO:
				printf("overwrite? : %s\n", cpinfo->dst);
				printf("y: Yes,  a: Yes (All),");

				if(type == 'd')
				{
					puts("  d: Yes (Only this folder),");
				}
				else
				{
					puts("");
				}

				printf("n:  No,  z:  No (All),");

				if(type == 'd')
				{
					puts("  o:  No (Only this folder),");
				}
				else
				{
					puts("");
				}

				printf(" : ");

				char stdin_buf[NAME_MAX];

				if(fgets(stdin_buf, NAME_MAX, stdin) == NULL)
				{
					fgets_is_null();
				}

				clean(stdin_buf, NAME_MAX);

				if(
					stdin_buf[0] != 'y' &&
					stdin_buf[0] != 'a' &&
					stdin_buf[0] != 'd' &&
					stdin_buf[0] != 'n' &&
					stdin_buf[0] != 'o' &&
					stdin_buf[0] != 'z'
				)
				{
					goto INPUT;
				}

				if((type != 'd') && ((stdin_buf[0] == 'd') || (stdin_buf[0] == 'o')))
				{
					goto INPUT;
				}

				/* このフォルダで○○の時にフォルダパスを記憶しておく変数なのでstatic必須 */
				static char *this_dir;

				switch(stdin_buf[0])
				{
				case 'y':	/* 上書き */
					check_write_file_or_create_link(oparg, sdir, cpinfo, cpdd);
					break;

				case 'a':	/* 全て上書き */
					check_write_file_or_create_link(oparg, sdir, cpinfo, cpdd);
					I = NOT;	/* 上書き確認モード無効 */
					/* WMODE = OVERWRITE; */
					/* ↑上書き確認モードでは最初からOVERWRITEになっている */
					break;

				case 'd':	/* このフォルダでのみ上書き */
					check_write_file_or_create_link(oparg, sdir, cpinfo, cpdd);
					iowt = THIS_DIR_YES;
					FREE_THIS_DIR
					MAKE_THIS_DIR
					break;

				case 'n':	/* 上書きしない */
					DO_NOT_OVERWRITE
					break;

				case 'o':	/* このフォルダでは上書きしない */
					DO_NOT_OVERWRITE
					iowt = THIS_DIR_NO;
					FREE_THIS_DIR
					MAKE_THIS_DIR
					break;

				case 'z':	/* 全て上書きしない */
					DO_NOT_OVERWRITE
					I = NOT;	/* 上書き確認モード無効 */
					WMODE = NO_OVERWRITE;	/* 非上書きモード */
					break;
				}
				break;

			case THIS_DIR_YES:
				{
					_Bool b = check_parent(this_dir, cpinfo->dst);

					if(b == true)
					{
						check_write_file_or_create_link(oparg, sdir, cpinfo, cpdd);
					}
					else
					{
						/* 再び上書き確認モードに */
						iowt = INFO;
						FREE_THIS_DIR
						goto INPUT;
					}
				}
				break;

			case THIS_DIR_NO:
				{
					_Bool b = check_parent(this_dir, cpinfo->dst);

					if(b == true)
					{
						DO_NOT_OVERWRITE
					}
					else
					{
						/* 再び上書き確認モードに */
						iowt = INFO;
						FREE_THIS_DIR
						goto INPUT;
					}
				}
				break;
			}
		}
		else
		{
			if(WMODE == NO_OVERWRITE)
			{
				DO_NOT_OVERWRITE
			}
			else
			{
				check_write_file_or_create_link(oparg, sdir, cpinfo, cpdd);
			}
		}
	}
}

/*******************************************************************************
*******************************************************************************/
static void check_write_file_or_create_link(const OPArg *oparg, const SDir *sdir, CPInfo *cpinfo, CPDD *cpdd)
{
	if(cpinfo->src_type == REGULAR)
	{
		read_write(oparg, cpinfo, cpdd);

		if(oparg->CHECK_MODE == BEFORE)
		{
			compare_verify_one(oparg, sdir, cpinfo, cpdd);
		}

		if((oparg->M == MOVE) && (cpinfo->dst_para == WRITE))
		{
			if((oparg->CHECK == NOT) || ((oparg->CHECK_MODE == BEFORE) && (cpinfo->hash_check == true)))
			{
				REMOVE_SOURCE_UNLINK
			}
		}
	}
	else
	{
		create_symboliclink(oparg, cpinfo, cpdd);

		if((oparg->M == MOVE) && (cpinfo->dst_para == WRITE) && (oparg->CHECK_MODE != AFTER))
		{
			REMOVE_SOURCE_UNLINK
		}
	}
}

/*******************************************************************************
*******************************************************************************/
static inline _Bool check_parent(const char *src, const char *dst)
{
	_Bool flag = false;

	for(;;)
	{
		if(*src == *dst)
		{
			if(*src == '\0')
			{
				flag = true;
				break;
			}

			src++;
			dst++;
		}
		else if((*src == '\0') && (*dst == G_DIR_SEPARATOR))
		{
			flag = true;
			break;
		}
		else 
		{
			break;
		}
	}

	return flag ? true : false;
}

/*******************************************************************************
*******************************************************************************/
static inline struct stat * make_directory(const OPArg *oparg, CPInfo *cpinfo, CPDD *cpdd)
{
	errno = 0;
	if(g_file_test(cpinfo->dst, G_FILE_TEST_EXISTS) == TRUE)
	{
		/* フォルダの作成は行わない */
		DO_NOT_OVERWRITE
	}
	else if(mkdir(cpinfo->dst, 0755) == -1)
	{
		cpinfo->dst_para = INITIAL;

		if((errno != EEXIST) && (errno != 0))
		{
			print_error("mkdir", __FILE__, __LINE__, cpdd);
			fprintf(stderr, "error : %s\n", cpinfo->dst);
		}
	}
	else
	{
		if(oparg->V == VERBOS)
		{
			printf("%s -> %s\n", cpinfo->src, cpinfo->dst);
			fflush(stdout);
		}

		cpinfo->dst_para = WRITE;

		if(cpdd->mk_dir_count <= ULLONG_MAX)
		{
			cpdd->mk_dir_count++;
		}
	}

	struct stat *stat_buf = xmalloc(sizeof(struct stat));

	if(stat(cpinfo->src, stat_buf) != 0)
	{
		print_error("stat", __FILE__, __LINE__, cpdd);
		fprintf(stderr, "error : %s\n", cpinfo->src);
		free(stat_buf);
		stat_buf = NULL;
	}

	return stat_buf;
}

/*******************************************************************************
*******************************************************************************/
#include <dirent.h>

#define TYPE_IS_LINK \
{\
	count_plus_link(cpinfo, cpdd);\
	check_overwrite(oparg, sdir, cpinfo, cpdd, 'd');\
	file_list = append_list((void *)cpinfo, file_list);\
}

#define TYPE_IS_FILE \
{\
	count_plus_file(cpinfo, cpdd);\
	cpinfo->copy_mode = cpinfo_tmp->copy_mode;\
	check_overwrite(oparg, sdir, cpinfo, cpdd, 'd');\
	file_list = append_list((void *)cpinfo, file_list);\
}

#define TYPE_IS_DIR \
{\
	count_plus_dir(cpinfo, cpdd);\
	cpinfo->copy_mode = cpinfo_tmp->copy_mode;\
	struct stat *stat_buf = make_directory(oparg, cpinfo, cpdd);\
	file_list = append_list((void *)cpinfo, file_list);\
	/* 再帰 */ \
	file_list = dir_traverse(oparg, sdir, file_list, cpinfo, cpdd);\
	if(cpinfo->dst_para == WRITE) {ch_time_and_mod_dir(cpinfo, stat_buf, cpdd); }\
	if((oparg->M == MOVE) && ((oparg->CHECK_MODE == BEFORE) || (oparg->CHECK_MODE == NOT))) { REMOVE_SOURCE_RMDIR }\
}

/*******************************************************************************
 * フォルダを走査
*******************************************************************************/
static VList * dir_traverse(const OPArg *oparg, const SDir *sdir, VList *file_list, CPInfo *cpinfo_tmp, CPDD *cpdd)
{
	DIR *dp = NULL;
	const char *src_dir = cpinfo_tmp->src;
	const int src_dir_len = strlen(src_dir);

	if((dp = opendir(src_dir)) != NULL)
	{
		struct dirent *entry = NULL;
		const char *d_name;
		const char *dst_dir = cpinfo_tmp->dst;
		const int dst_dir_len = strlen(dst_dir);

		while((entry = readdir(dp)) != NULL)
		{
			d_name = entry->d_name;

			/* strcmpは等しければ0を返す（0は偽） */
			if(strcmp(d_name, ".") && strcmp(d_name, ".."))
			{
				const int d_name_len = strlen(d_name);

				CPInfo *cpinfo = malloc_cpinfo();

				/* +2は、 \0 と / の分 */
				cpinfo->src = xmalloc(src_dir_len + d_name_len + 2);
				cpinfo->dst = xmalloc(dst_dir_len + d_name_len + 2);
				cpinfo->dstfs = cpinfo_tmp->dstfs;
				cpinfo->copy_mode = cpinfo_tmp->copy_mode;
				/*
				 * sprintfは便利だってLeptonさんが言ってた。
				 * printf系の関数は遅い？知るか。
				 */
				sprintf(cpinfo->src, "%s%s%s", src_dir, G_DIR_SEPARATOR_S, d_name);
				sprintf(cpinfo->dst, "%s%s%s", dst_dir, G_DIR_SEPARATOR_S, d_name);

				switch(entry->d_type)
				{
				case DT_LNK:
					TYPE_IS_LINK
					break;

				case DT_REG:
					TYPE_IS_FILE
					break;

				case DT_DIR:
					TYPE_IS_DIR
					break;

				default:
					if(g_file_test(cpinfo->src, G_FILE_TEST_IS_SYMLINK) == TRUE)
					{
						TYPE_IS_LINK
					}
					else if(g_file_test(cpinfo->src, G_FILE_TEST_IS_REGULAR) == TRUE)
					{
						TYPE_IS_FILE
					}
					else if(g_file_test(cpinfo->src, G_FILE_TEST_IS_DIR) == TRUE)
					{
						TYPE_IS_DIR
					}
					else
					{
						fprintf(stderr, "skip : %s\n", cpinfo->src);
					}
					break;
				}
			}
		}

		if(closedir(dp) == -1)
		{
			print_error("closedir", __FILE__, __LINE__, cpdd);
			fprintf(stderr, "error : %s\n", src_dir);
		}
	}
	else
	{
		print_error("opendir", __FILE__, __LINE__, cpdd);
		fprintf(stderr, "error : %s\n", src_dir);
	}

	/* フォルダ削除用 */
	CPInfo *cpinfo_dir = malloc_cpinfo();
	cpinfo_dir->src = xmalloc(src_dir_len + 1);
	strcpy(cpinfo_dir->src, cpinfo_tmp->src);
	cpinfo_dir->src_type = cpinfo_tmp->src_type;
	cpinfo_dir->dst_para = cpinfo_tmp->dst_para;
	cpinfo_dir->src_para = cpinfo_tmp->src_para;
	cpinfo_dir->is_rmdir = true;
	file_list = append_list((void *)cpinfo_dir, file_list);

	return file_list;
}

/*******************************************************************************
 * 文字列の最後がパス区切り文字だった場合\0にする
*******************************************************************************/
static inline void delete_last_slash(const int len, char tmp[])
{
	if(tmp[len - 1] == G_DIR_SEPARATOR)
	{
		tmp[len - 1] = '\0';
	}
}

/******************************************************************************
*******************************************************************************/
static inline CPInfo * malloc_cpinfo(void)
{
	CPInfo *cpinfo = xmalloc(sizeof(CPInfo));
	cpinfo->src = NULL;
	cpinfo->dst = NULL;
	cpinfo->dstfs = NULL;
	cpinfo->copy_mode = SAME;
	cpinfo->dst_para = INITIAL;
	cpinfo->src_para = DEFAULT;
	cpinfo->hash_check = false;
	cpinfo->file_test = false;
	cpinfo->src_md5 = NULL;
	cpinfo->src_size = 0;
	cpinfo->dst_size = 0;
	cpinfo->is_rmdir = false;

	return cpinfo;
}

/*******************************************************************************
 * コピー元ファイル名とコピー先パスを連結する。
 * g_path_get_basename と g_build_path を同時にやる感じ。
*******************************************************************************/
static inline char * get_dst_plus_srcname(const char *src, const char *dst)
{
	char sname[PATH_MAX];
	char *return_name;
	int src_len = strlen(src);
	int dst_len = strlen(dst);
	_Bool flag = false;

	for(int i = src_len - 1; i >= 0; i--)
	{
		/*
		 * 最後のパス区切り文字からコピーするので、
		 * fnameは以下となる。
		 * 
		 * /ファイル名\0
		 */
		if(src[i] == G_DIR_SEPARATOR)
		{
			flag = true;
			int j = 0;

			for(;;)
			{
				sname[j] = src[i];

				if(src[i] == '\0')
				{
					goto BASENAME_END;
				}

				i++;
				j++;
			}
		}
	}

BASENAME_END:

	if(flag == true)
	{
		int name_len = strlen(sname);
		/* +1 はヌル文字（\0）の分 */
		return_name = xmalloc(dst_len + name_len + 1);
		memcpy(return_name, dst, dst_len + 1);
		strcat(return_name, sname);
	}
	else
	{
		/*
		 * 2011年3月19日
		 * メモリ破壊のバグぅうわあああ。
		 * append_list内のxmalloc内で、malloc(): memory corruption
		 * とか出てアボートするバグを直せた。
		 * return_name = xmalloc(src_len + dst_len + 1);
		 * ↑こいつが原因だった。G_DIR_SEPARATOR_Sのことを忘れてた。
		 * 多分これで大丈夫。
		 */
		return_name = xmalloc(src_len + dst_len + (sizeof(G_DIR_SEPARATOR_S) + 1));
		memcpy(return_name, dst, dst_len + 1);
		strcat(return_name, G_DIR_SEPARATOR_S);
		strcat(return_name, src);
	}

	return return_name;
}

/******************************************************************************
*******************************************************************************/
static inline void count_plus_link(CPInfo *cpinfo, CPDD *cpdd)
{
	cpinfo->src_type = SYMBOLICLINK;

	if(cpdd->link_count <= ULLONG_MAX)
	{
		cpdd->link_count++;
	}
}

/******************************************************************************
*******************************************************************************/
static inline void count_plus_file(CPInfo *cpinfo, CPDD *cpdd)
{
	cpinfo->src_type = REGULAR;

	if(cpdd->file_count <= ULLONG_MAX)
	{
		cpdd->file_count++;
	}
}

/******************************************************************************
*******************************************************************************/
static inline void count_plus_dir(CPInfo *cpinfo, CPDD *cpdd)
{
	cpinfo->src_type = DIRECTORY;

	if(cpdd->dir_count <= ULLONG_MAX)
	{
		cpdd->dir_count++;
	}
}

/*******************************************************************************
 * 入力バッファのクリア
*******************************************************************************/
static inline void clean(const char stdin_buf[], const int INPUT_LEN)
{
	/* バッファチェック */
	_Bool flag = false;
	for(int i = 0; i < INPUT_LEN; i++)
	{
		if(stdin_buf[i] == '\n')
		{
			flag = true;
			break;
		}
	}
	/* 入力バッファの掃除 */
	if(flag == false)
	{
		while(getchar() != '\n');
	}
}

/*******************************************************************************
 * fgetsの戻り値がNULLだった場合
*******************************************************************************/
static inline void fgets_is_null(void)
{
	fprintf(stderr, "fgetsの戻り値がNULLです\n");
	exit(EXIT_FAILURE);
}

/******************************************************************************
*******************************************************************************/
#define PATH_CHECK \
const char *mp = tmp->mount_point;\
_Bool flag = false;\
\
for(;;)\
{\
	if(*path == *mp)\
	{\
		path++;\
		mp++;\
	}\
	else \
	{\
		break;\
	}\
\
	if(*mp == '\0')\
	{\
		if((*path == G_DIR_SEPARATOR) || (*path == '\0'))\
		{\
			flag = true;\
		}\
		break;\
	}\
}

/******************************************************************************
 * ドライブが同一かどうか調べる
*******************************************************************************/
static inline void set_same_or_different(const VList *drive_list, const char *src, const char *dst, const OPArg *oparg, CPInfo *cpinfo)
{
	/*
	 * キャストしないと警告が出る
	 * warning: initialization discards qualifiers from pointer target type
	 */
	VList *tmplist = (VList *)drive_list;
	char *srcdev = NULL;
	char *dstdev = NULL;

	for(;;)
	{
		DPath *tmp = (DPath *)tmplist->data;

		/* /の処理 */
		if((tmp->mount_point[1] == '\0') && (tmp->mount_point[0] == G_DIR_SEPARATOR))
		{
			if(srcdev == NULL)
			{
				srcdev = tmp->device;
			}

			if(dstdev == NULL)
			{
				dstdev = tmp->device;
				cpinfo->dstfs = tmp->fs;
			}

			goto DEV_CHECK_SKIP;
		}

		{
			const char *path = src;

			PATH_CHECK

			if(flag == true)
			{
				srcdev = tmp->device;
			}
		}

		{
			const char *path = dst;

			PATH_CHECK

			if(flag == true)
			{
				dstdev = tmp->device;
				cpinfo->dstfs = tmp->fs;
			}
		}

DEV_CHECK_SKIP:
;
		if(tmplist->next != NULL)
		{
			tmplist = tmplist->next;
		}
		else
		{
			break;
		}
	}

	if((srcdev != NULL) && (dstdev != NULL))
	{
		if(strcmp(srcdev, dstdev) == 0)
		{
			cpinfo->copy_mode = SAME;
		}
		else
		{
			cpinfo->copy_mode = DIFFERENT;
		}
	}
	else
	{
		cpinfo->copy_mode = SAME;
	}

	switch(oparg->THREAD_MODE)
	{
	case DRIVE_DIFFERENT:
		cpinfo->copy_mode = DIFFERENT;
		break;

	case DRIVE_AUTO:
		break;

	default:
		cpinfo->copy_mode = SAME;
		break;
	}
}
